import re

from django.core.exceptions import ValidationError
from django.db import models
from django.utils.translation import ugettext_lazy as _

from datetime import timedelta
from dateutil.relativedelta import relativedelta

# This is not quite ISO8601, as it allows the SQL/Postgres extension
# of allowing a minus sign in the values, and you can mix weeks with
# other units (which ISO doesn't allow).
iso8601_duration_re = re.compile(
    r'P'
    r'(?:(?P<years>-?\d+(?:\.\d+)?)Y)?'
    r'(?:(?P<months>-?\d+(?:\.\d+)?)M)?'
    r'(?:(?P<weeks>-?\d+(?:\.\d+)?)W)?'
    r'(?:(?P<days>-?\d+(?:\.\d+)?)D)?'
    r'(?:T'
    r'(?:(?P<hours>-?\d+(?:\.\d+)?)H)?'
    r'(?:(?P<minutes>-?\d+(?:\.\d+)?)M)?'
    r'(?:(?P<seconds>-?\d+(?:\.\d+)?)S)?'
    r')?'
    r'$'
)

# Parse ISO8601 timespec
def parse_relativedelta(str):
	m = iso8601_duration_re.match(str)
	if m:
		args = {}
		for k, v in m.groupdict().items():
			if v is  None:
				args[k] = 0
			elif '.' in v:
				args[k] = float(v)
			else:
				args[k] = int(v)
		return relativedelta(**args).normalized() if m else None

	raise ValueError('Not a valid (extended) ISO8601 interval specification')


# Format ISO8601 timespec
def format_relativedelta(relativedelta):
	result_big = ''
	# TODO: We could always include all components, but that's kind of
	# ugly, since one second would be formatted as 'P0Y0M0W0DT0M1S'
	if relativedelta.years:
		result_big += '{}Y'.format(relativedelta.years)
	if relativedelta.months:
		result_big += '{}M'.format(relativedelta.months)
	if relativedelta.days:
		result_big += '{}D'.format(relativedelta.days)

	result_small = ''
	if relativedelta.hours:
		result_small += '{}H'.format(relativedelta.hours)
	if relativedelta.minutes:
		result_small += '{}M'.format(relativedelta.minutes)
	# Microseconds is allowed here as a convenience, the user may have
	# used normalized(), which can result in microseconds
	if relativedelta.seconds:
		seconds = relativedelta.seconds
		if relativedelta.microseconds:
			seconds += relativedelta.microseconds / 1000000.0
		result_small += '{}S'.format(seconds)

	if len(result_small) > 0:
		return 'P{}T{}'.format(result_big, result_small)
	elif len(result_big) == 0:
		return 'P0D' # Doesn't matter much what field is zero, but just 'P' is invalid syntax, and so is ''
	else:
		return 'P{}'.format(result_big)



class RelativeDeltaField(models.Field):
	"""Stores dateutil.relativedelta.relativedelta objects.

	Uses INTERVAL on PostgreSQL.
	"""
	empty_strings_allowed = False
	default_error_messages = {
		'invalid': _("'%(value)s' value has an invalid format. It must be in "
					 "ISO8601 interval format.")
	}
	description = _("RelativeDelta")


	def db_type(self, connection):
		if connection.settings_dict['ENGINE'] in ('django.db.backends.postgresql_psycopg2', 'django.db.backends.postgresql', 'django.contrib.gis.db.backends.postgis'):
			return 'interval'
		else:
			raise ValueError(_('RelativeDeltaField only supports PostgreSQL for storage'))


	def to_python(self, value):
		if value is None:
			return value
		elif isinstance(value, relativedelta):
			return value.normalized()
		elif isinstance(value, timedelta):
			return (relativedelta() + value).normalized()

		try:
			return parse_relativedelta(value)
		except (ValueError, TypeError):
			raise ValidationError(
				self.error_messages['invalid'],
				code='invalid',
				params={'value': value},
			)


	def get_db_prep_value(self, value, connection, prepared=False):
		if value is None:
			return value
		else:
			return format_relativedelta(self.to_python(value))


	# This is a bit of a mindfuck.  We have to cast the output field
	# as text to bypass the standard deserialisation of PsycoPg2 to
	# datetime.timedelta, which loses information.  We then parse it
	# ourselves in convert_relativedeltafield_value().
	#
	# We make it easier for ourselves by doing some formatting here,
	# so that we don't need to rely on weird detection logic for the
	# current value of IntervalStyle (PsycoPg2 actually gets this
	# wrong; it only checks / sets DateStyle, but not IntervalStyle)
	#
	# We can't simply replace or remove PsycoPg2's parser, because
	# that would mess with any existing Django DurationFields, since
	# Django assumes PsycoPg2 returns pre-parsed datetime.timedeltas.
	def select_format(self, compiler, sql, params):
		fmt = 'to_char(%s, \'PYYYY"Y"MM"M"DD"DT"HH24"H"MI"M"SS.US"S"\')' % sql
		return fmt, params


	def get_db_converters(self, connection):
		return [self.convert_relativedeltafield_value]


	def convert_relativedeltafield_value(self, value, expression, connection, context):
		if value is not None:
			return parse_relativedelta(value)


	def value_to_string(self, obj):
		val = self.value_from_object(obj)
		return '' if val is None else format_relativedelta(val)



from django import setup
from django.conf import settings

settings.configure(**{
	'DEBUG': True,
	'DATABASES': {
		'default': {
			'ENGINE': 'django.db.backends.postgresql_psycopg2',
			'NAME': 'relativedelta-test',
		},
	},
	'MIDDLEWARE': [
		# TODO: Try to reduce the set of absolutely required middlewares
		'request_id.middleware.RequestIdMiddleware',
		'django.contrib.sessions.middleware.SessionMiddleware',
		'django.contrib.auth.middleware.AuthenticationMiddleware',
	],
	'INSTALLED_APPS': [
		'relativedeltafield',
		'relativedeltafield.tests.testapp',
	],
	'MIGRATION_MODULES': {
		'testapp': None,
		'relativedeltafield': None,
	},
	'USE_TZ': True,
	#'ROOT_URLCONF': 'relativedeltafield.tests.testapp.urls',
	'LOGGING': {
		'version': 1,
		'handlers': {
			'console': {
				'level': 'DEBUG',
				'class': 'logging.StreamHandler',
			},
		},
		'loggers': {
			# We override only this one to avoid logspam
			# while running tests.  Django warnings are
			# stil shown.
			'django-relativedelta': {
				'handlers': ['console'],
				'level': 'ERROR',
			},
		}
	}
})

setup()

# Do the dance to ensure the models are synched to the DB.
# This saves us from having to include migrations
from django.core.management.commands.migrate import Command as MigrationCommand
from django.db import connections
from django.db.migrations.executor import MigrationExecutor

# This is oh so hacky....
cmd = MigrationCommand()
cmd.verbosity = 0
connection = connections['default']
executor = MigrationExecutor(connection)
cmd.sync_apps(connection, executor.loader.unmigrated_apps)



from django.core.exceptions import ValidationError
from django.test import TestCase

from relativedeltafield import RelativeDeltaField
from .testapp.models import Interval

from datetime import timedelta
from dateutil.relativedelta import relativedelta

class RelativeDeltaFieldTest(TestCase):
	def setUp(self):
		Interval.objects.all().delete()

	def assertStrictEqual(self, a, b):
		self.assertEqual(a, b)
		self.assertEqual(type(a), type(b))


	def test_basic_value_survives_db_roundtrip(self):
		input_value = relativedelta(years=2,months=3,days=4,hours=5,minutes=52,seconds=30,microseconds=5)
		obj = Interval(value=input_value)
		obj.save()

		obj.refresh_from_db()
		self.assertStrictEqual(input_value, obj.value)


	def test_empty_value_survives_db_roundtrip(self):
		obj = Interval(value=relativedelta())
		obj.save()

		obj.refresh_from_db()
		self.assertStrictEqual(relativedelta(), obj.value)


	def test_each_separate_value_survives_db_roundtrip(self):
		values = {
			'years': 501,
			'months': 10,
			'days': 2,
			'hours': 1,
			'minutes': 52,
			'seconds': 12,
		}

		for k in values:
			input_value = relativedelta(**{k: values[k]})
			obj = Interval(value=input_value)
			obj.save()

			obj.refresh_from_db()
			# Put the object in a dict to get descriptive output on failure
			self.assertEqual({k: input_value}, {k: obj.value})
			self.assertEqual({k: int}, {k: type(getattr(obj.value, k))})


	# See issue #2, we should not serialize weeks separately, because
	# it is a derived value.
	def test_weeks_value_survives_db_roundtrip_as_days(self):
		input_value = relativedelta(weeks=2,days=1)
		obj = Interval(value=input_value)
		obj.save()

		obj.refresh_from_db()
		self.assertStrictEqual(15, obj.value.days)
		self.assertStrictEqual(2, obj.value.weeks)


	def test_none_value_also_survives_db_roundtrip(self):
		obj = Interval(value=None)
		obj.save()

		obj.refresh_from_db()
		self.assertIsNone(obj.value)


	def test_none_value_survives_full_clean(self):
		obj = Interval(value=None)
		obj.full_clean()
		self.assertIsNone(obj.value)


	# Specific check, because to_python doesn't get called by save()
	# or full_clean() when the value is None (but other things might)
	# This is a regression test for a bug where we'd call normalized()
	# even on None values.
	def test_none_value_survives_to_python(self):
		self.assertIsNone(RelativeDeltaField().to_python(None))


	def test_value_is_normalized_on_full_clean(self):
		input_value = relativedelta(years=1,months=3,weeks=1,days=4.5,hours=5,minutes=70.5,seconds=80.100005,microseconds=5)
		obj = Interval(value=input_value)
		obj.full_clean()

		self.assertNotEqual(input_value, obj.value)
		self.assertStrictEqual(input_value.normalized(), obj.value)

		# Quick sanity check to ensure the input isn't mutated.
		# Take into account that weeks are added to days though!
		self.assertStrictEqual(11.5, input_value.days)
		self.assertStrictEqual(1, input_value.weeks)

		# Check that the values are normalized
		self.assertStrictEqual(1, obj.value.years)
		self.assertStrictEqual(3, obj.value.months)
		self.assertStrictEqual(11, obj.value.days)
		self.assertStrictEqual(18, obj.value.hours)
		self.assertStrictEqual(11, obj.value.minutes)
		self.assertStrictEqual(50, obj.value.seconds)
		self.assertStrictEqual(100010, obj.value.microseconds)

		# Derived value, from the number of days (see #2); but it's no
		# longer converted to a floating-point number in newer
		# versions of relativedelta....
		self.assertStrictEqual(1, obj.value.weeks)


	def test_weeks_value_is_derived_as_int_when_normalizing_on_full_clean(self):
		input_value = relativedelta(years=1,months=3,weeks=1,days=11.5,hours=5,minutes=70.5,seconds=80.100005,microseconds=5)
		obj = Interval(value=input_value)
		obj.full_clean()

		self.assertNotEqual(input_value, obj.value)
		self.assertStrictEqual(input_value.normalized(), obj.value)

		# Quick sanity check to ensure the input isn't mutated.
		# Take into account that weeks are added to days though!
		self.assertStrictEqual(18.5, input_value.days)
		self.assertStrictEqual(2, input_value.weeks)

		# Derived value, from the number of days (see #2)
		self.assertStrictEqual(2, obj.value.weeks)


	def test_string_input(self):
		obj = Interval(value='P1Y3M1W4.5DT5H70.5M80.10001S')
		obj.full_clean()

		self.assertIsInstance(obj.value, relativedelta)

		# Check that the values are normalized
		self.assertStrictEqual(1, obj.value.years)
		self.assertStrictEqual(3, obj.value.months)
		self.assertStrictEqual(11, obj.value.days)
		self.assertStrictEqual(18, obj.value.hours)
		self.assertStrictEqual(11, obj.value.minutes)
		self.assertStrictEqual(50, obj.value.seconds)
		self.assertStrictEqual(100010, obj.value.microseconds)

		# Derived value, from the number of days (see #2)
		self.assertStrictEqual(1, obj.value.weeks)


	def test_invalid_string_inputs_raise_validation_error(self):
		obj = Interval()

		obj.value = 'blabla'
		with self.assertRaises(ValidationError) as cm:
			obj.full_clean()
		self.assertEqual(set(['value']), set(cm.exception.message_dict.keys()))

		obj.value = 'P1.5M' # not allowed by relativedelta because it is supposedly ambiguous
		with self.assertRaises(ValidationError) as cm:
			obj.full_clean()
		self.assertEqual(set(['value']), set(cm.exception.message_dict.keys()))

		obj.value = 'P1M' # Check that the error is cleared when made valid again
		obj.full_clean()


	def test_invalid_objects_raise_validation_errors(self):
		obj = Interval()

		obj.value = True
		with self.assertRaises(ValidationError) as cm:
			obj.full_clean()
		self.assertEqual(set(['value']), set(cm.exception.message_dict.keys()))

		obj.value = 1
		with self.assertRaises(ValidationError) as cm:
			obj.full_clean()
		self.assertEqual(set(['value']), set(cm.exception.message_dict.keys()))

		obj.value = 'P1M' # Check that the error is cleared when made valid again
		obj.full_clean()


	def test_timedelta_input(self):
		td = timedelta(weeks=1,days=4.5,hours=5,minutes=70.5,seconds=80.100005,microseconds=5)
		obj = Interval(value=td)
		obj.full_clean()

		self.assertIsInstance(obj.value, relativedelta)

		# Check that the values are normalized
		self.assertStrictEqual(0, obj.value.years)
		self.assertStrictEqual(0, obj.value.months)
		self.assertStrictEqual(11, obj.value.days)
		self.assertStrictEqual(18, obj.value.hours)
		self.assertStrictEqual(11, obj.value.minutes)
		self.assertStrictEqual(50, obj.value.seconds)
		self.assertStrictEqual(100010, obj.value.microseconds)

		# Derived value, from the number of days (see #2)
		self.assertStrictEqual(1, obj.value.weeks)


	def test_filtering_works(self):
		obj1 = Interval(value='P1Y3M1W4.5DT5H70.5M80.10001S')
		obj1.save()

		obj2 = Interval(value='P12D')
		obj2.save()

		q = Interval.objects.filter(value__gt='P1Y')
		self.assertEqual(1, q.count())

		q = Interval.objects.filter(value__lt='P2Y')
		self.assertEqual(2, q.count())



#! /usr/bin/env python3

import os
from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme:
	README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
	name='django-relativedelta',
	version='1.0.5',
	packages=find_packages(),
	include_package_data=True,
	license='MIT License',
	description='Django alternative to DurationField using dateutil.relativedelta',
	long_description=README,
	url='https://github.com/CodeYellowBV/django-relativedelta',
	author='Peter Bex',
	author_email='peter@codeyellow.nl',
	test_suite='relativedeltafield.tests',
	classifiers=[
		'Environment :: Web Environment',
		'Framework :: Django',
		'Framework :: Django :: 1.10',
		'Framework :: Django :: 1.11',
		'Intended Audience :: Developers',
		'License :: OSI Approved :: MIT License',
		'Operating System :: OS Independent',
		'Programming Language :: Python',
		'Programming Language :: Python :: 3',
		'Programming Language :: Python :: 3.4',
		'Programming Language :: Python :: 3.5',
		'Topic :: Database',
		'Topic :: Internet :: WWW/HTTP',
		'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
	],
	install_requires=[
		'Django >= 1.10',
		'python-dateutil >= 2.6.0',
                # Disabled (for now?) due to https://github.com/CodeYellowBV/django-relativedelta/issues/6
                # We never import it directly from the code, so that is okay
		#'psycopg2 >= 2.7.0',
	],
	tests_require=[
                # The tests do need psycopg2
		'psycopg2 >= 2.7.0',
        ],
)
